#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Multiple entities entity and resolver."""

from __future__ import annotations

import dataclasses
import json
from collections import defaultdict
from typing import TYPE_CHECKING, Any, cast, override

from invenio_records_resources.references.entity_resolvers import EntityProxy
from invenio_records_resources.references.entity_resolvers.base import EntityResolver
from invenio_records_resources.services.records.results import FieldsResolver
from invenio_requests.resolvers.registry import ResolverRegistry
from invenio_requests.services.results import EntityResolverExpandableField

if TYPE_CHECKING:
    from collections.abc import Mapping

    from flask_principal import Identity, ItemNeed, Need


@dataclasses.dataclass
class MultipleEntitiesEntity:
    """Entity representing multiple entities.

    Implementation note: entities is intentionally list of EntityProxy and not list of resolved entities.
    The reason is that these are used to be converted to entity references and having proxy here makes it easier.
    """

    entities: list[EntityProxy]

    @classmethod
    def create_id(cls, entity_references: list[Mapping[str, str]]) -> str:
        """Create id from entity references."""
        entity_references.sort(key=lambda x: (next(iter(x.keys())), next(iter(x.values()))))
        return json.dumps(entity_references, sort_keys=True)

    @property
    def id(self) -> str:
        """Return id of the entity."""
        ref_dict_list = [entity.reference_dict for entity in self.entities]
        ref_dict_list.sort(key=lambda x: (next(iter(x.keys())), next(iter(x.values()))))
        return json.dumps(ref_dict_list, sort_keys=True)


class MultipleEntitiesProxy(EntityProxy):
    """Proxy for multiple-entities entity."""

    def _resolve(self) -> MultipleEntitiesEntity:
        """Resolve the entity reference into entity."""
        values = json.loads(self._parse_ref_dict_id())
        return MultipleEntitiesEntity(
            entities=[
                cast(
                    "EntityProxy",
                    ResolverRegistry.resolve_entity_proxy(ref, raise_=True),
                )
                for ref in values
            ]
        )

    @override
    def get_needs(self, ctx: dict | None = None) -> list[Need | ItemNeed]:
        """Get needs that the entity generate."""
        ret: list[Need | ItemNeed] = []
        entity = self._entity or self._resolve()
        for subentity_proxy in entity.entities:
            ret.extend(subentity_proxy.get_needs(ctx) or [])
        return ret

    @override
    def pick_resolved_fields(self, identity: Identity, resolved_dict: dict[str, Any]) -> dict[str, Any]:
        """Pick resolved fields for serialization of the entity to json."""
        entity_refs = json.loads(resolved_dict["id"])
        field_keys = []
        hit: dict[str, dict[str, dict[str, str]]] = defaultdict(dict)
        for entity_ref in entity_refs:
            type_ = next(iter(entity_ref.keys()))
            id_ = next(iter(entity_ref.values()))
            field_keys.append(f"{type_}.{id_}")
            hit[type_] |= {id_: {type_: id_}}
        fr = FieldsResolver([EntityResolverExpandableField(field_key) for field_key in field_keys])
        fr.resolve(identity, [hit])
        return fr.expand(identity, hit)  # type: ignore[no-any-return]


class MultipleEntitiesResolver(EntityResolver):
    """A resolver that resolves multiple entities entity."""

    # TODO: move as constant to MultipleEntitiesEntity? Would be similar to AutoApprove
    type_id = "multiple"

    def __init__(self) -> None:
        """Initialize the resolver."""
        super().__init__("multiple")

    @override
    def matches_reference_dict(self, ref_dict: dict) -> bool:
        """Check if the reference dictionary can be resolved by this resolver."""
        return cast("bool", self._parse_ref_dict_type(ref_dict) == self.type_id)

    @override
    def _reference_entity(self, entity: MultipleEntitiesEntity) -> dict[str, str]:
        """Return a reference dictionary for the entity."""
        return {self.type_id: entity.id}

    @override
    def matches_entity(self, entity: Any) -> bool:
        """Check if the entity can be serialized to a reference by this resolver."""
        return isinstance(entity, MultipleEntitiesEntity)

    @override
    def _get_entity_proxy(self, ref_dict: dict) -> MultipleEntitiesProxy:
        """Get the entity proxy for the reference dictionary.

        Note: the proxy is returned to ensure the entity is loaded lazily, when needed.
        :param ref_dict: Reference dictionary.
        """
        return MultipleEntitiesProxy(self, ref_dict)



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-workflows (see http://github.com/oarepo/oarepo-workflows).
#
# oarepo-workflows is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#

"""Module containing preset for adding workflow functionality to parent records.

This module defines a preset that adds workflow capabilities to parent records
in the Invenio model system by adding a workflow system field.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from oarepo_model.customizations import (
    Customization,
    PrependMixin,
)
from oarepo_model.presets import Preset

from oarepo_workflows.records.systemfields.workflow import WorkflowField

if TYPE_CHECKING:
    from collections.abc import Generator

    from oarepo_model.builder import InvenioModelBuilder
    from oarepo_model.model import InvenioModel


class WorkflowsParentRecordPreset(Preset):
    """Preset that modifies a ParentRecord class."""

    modifies = ("ParentRecord",)

    @override
    def apply(
        self,
        builder: InvenioModelBuilder,
        model: InvenioModel,
        dependencies: dict[str, Any],
    ) -> Generator[Customization]:
        class WorkflowsParentRecordMixin:
            """Base class for parent records in the model."""

            workflow = WorkflowField()

        yield PrependMixin(
            "ParentRecord",
            WorkflowsParentRecordMixin,
        )



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-workflows (see http://github.com/oarepo/oarepo-workflows).
#
# oarepo-workflows is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""OARepo workflows model package.

This module provides database models and system fields for implementing
workflow functionality in OARepo records, including workflow states,
transitions and metadata handling capabilities.
"""

from __future__ import annotations



#
# Copyright (C) 2025 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Default configuration for oarepo-workflows to be initialized at invenio_config.module entrypoint."""

from __future__ import annotations

from oarepo_workflows.requests.permissions import (
    CreatorsFromWorkflowRequestsPermissionPolicy,
)

REQUESTS_PERMISSION_POLICY = CreatorsFromWorkflowRequestsPermissionPolicy



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Events for workflow requests."""

from __future__ import annotations

import dataclasses
from functools import cached_property
from typing import TYPE_CHECKING, NoReturn

from oarepo_workflows.errors import EventTypeNotInWorkflowError
from oarepo_workflows.requests.generators.multiple_entities import (
    MultipleEntitiesGenerator,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    from invenio_records_permissions.generators import Generator as InvenioGenerator
    from oarepo_runtime.services.generators import Generator


@dataclasses.dataclass
class WorkflowEvent:
    """Class representing a workflow event."""

    submitters: Sequence[InvenioGenerator]
    """List of submitters to be used for the event.

       The generators supply needs. The user must have at least one of the needs
       to be able to create a workflow event.
    """

    @cached_property
    def submitter_generator(self) -> Generator:
        """Return the requesters as a single requester generator."""
        return MultipleEntitiesGenerator(self.submitters)


class WorkflowEvents(dict[str, WorkflowEvent]):
    """Class representing a collection of workflow events."""

    def __missing__(self, key: str) -> NoReturn:
        """Raise EventTypeNotInWorkflowError when a key is not found."""
        raise EventTypeNotInWorkflowError(key)



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-workflows (see http://github.com/oarepo/oarepo-workflows).
#
# oarepo-workflows is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#

"""Module providing workflow functionality for record schema.

This module defines a preset for extending record marshmallow schemas with workflow-related
fields, including state and state timestamp.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

import marshmallow as ma
from marshmallow_utils.fields import EDTFDateTimeString
from oarepo_model.customizations import Customization, PrependMixin
from oarepo_model.presets import Preset

if TYPE_CHECKING:
    from collections.abc import Generator

    from oarepo_model.builder import InvenioModelBuilder
    from oarepo_model.model import InvenioModel


class WorkflowsRecordSchemaPreset(Preset):
    """Preset for record marshmallow schema."""

    modifies = ("RecordSchema",)

    @override
    def apply(
        self,
        builder: InvenioModelBuilder,
        model: InvenioModel,
        dependencies: dict[str, Any],
    ) -> Generator[Customization]:
        class StateSchemaMixin:
            state = ma.fields.String(dump_only=True)
            state_timestamp = EDTFDateTimeString(dump_only=True)

        yield PrependMixin("RecordSchema", StateSchemaMixin)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Definition of workflow permissions."""

from __future__ import annotations

from typing import Any

from invenio_rdm_records.services.generators import IfRecordDeleted, RecordOwners
from invenio_records_permissions import RecordPermissionPolicy
from invenio_records_permissions.generators import (
    AuthenticatedUser,
    Disable,
    SystemProcess,
)
from invenio_users_resources.services.permissions import UserManager

from .generators import IfInState, SameAs


class DefaultWorkflowPermissions(RecordPermissionPolicy):
    """Base class for workflow permissions, subclass from it and put the result to Workflow constructor.

    Example:
        class MyWorkflowPermissions(DefaultWorkflowPermissions):
            can_read = [AnyUser()]
    in invenio.cfg
    WORKFLOWS = {
        'default': Workflow(
            permission_policy_cls = MyWorkflowPermissions, ...
        )
    }

    """

    files_edit = (
        IfInState("draft", [RecordOwners()]),
        IfInState("published", [Disable()]),
    )

    system_process = SystemProcess()

    def __init__(self, action_name: str, **over: Any) -> None:
        """Initialize the workflow permissions."""
        if hasattr(
            self, f"can_{action_name}"
        ):  # edge case in FromRecordWorkflow; request action doesn't have to be defined if it's not in workflow
            can = getattr(self, f"can_{action_name}")
            if self.system_process not in can:
                can = (*can, self.system_process)
                setattr(self.__class__, f"can_{action_name}", can)
        over["policy"] = self
        super().__init__(action_name, **over)

    can_read = (
        IfInState("draft", [RecordOwners()]),
        IfInState("published", [AuthenticatedUser()]),
    )

    # from RDM
    can_read_deleted = (IfRecordDeleted(then_=[UserManager, SystemProcess()], else_=[SameAs("can_read")]),)
    can_update = (IfInState("draft", [RecordOwners()]),)
    can_delete = (IfInState("draft", [RecordOwners()]),)
    can_create = (AuthenticatedUser(),)
    can_publish = (RecordOwners(),)
    can_new_version = (RecordOwners(),)
    can_edit = (RecordOwners(),)

    can_create_files = (SameAs("files_edit"),)
    can_set_content_files = (SameAs("files_edit"),)
    can_commit_files = (SameAs("files_edit"),)
    can_update_files = (SameAs("files_edit"),)
    can_delete_files = (SameAs("files_edit"),)
    can_read_files = (SameAs("can_read"),)
    can_get_content_files = (SameAs("can_read"),)
    can_list_files = (SameAs("can_read"),)
    can_manage_files = (Disable(),)

    can_read_draft = (SameAs("can_read"),)
    can_update_draft = (SameAs("can_update"),)
    can_delete_draft = (SameAs("can_delete"),)

    # RDM records permissions
    can_bulk_add = (Disable(),)

    # draft files
    can_draft_commit_files = (SameAs("can_update_draft"),)
    can_draft_delete_files = (SameAs("can_update_draft"),)
    can_draft_get_content_files = (SameAs("can_read_draft"),)
    can_draft_read_files = (SameAs("can_read_draft"),)
    can_draft_set_content_files = (SameAs("can_update_draft"),)
    can_draft_update_files = (SameAs("can_update_draft"),)
    can_draft_create_files = (SameAs("files_edit"),)

    # draft media files
    can_draft_media_commit_files = (SameAs("can_update_draft"),)
    can_draft_media_create_files = (SameAs("can_update_draft"),)
    can_draft_media_delete_files = (SameAs("can_update_draft"),)
    can_draft_media_get_content_files = (SameAs("can_read_draft"),)
    can_draft_media_read_files = (SameAs("can_read_draft"),)
    can_draft_media_set_content_files = (SameAs("can_update_draft"),)
    can_draft_media_update_files = (SameAs("can_update_draft"),)

    # published media files
    can_media_commit_files = (Disable(),)
    can_media_create_files = (Disable(),)
    can_media_delete_files = (Disable(),)
    can_media_get_content_files = (SameAs("can_read_files"),)
    can_media_read_deleted_files = (Disable(),)
    can_media_read_files = (SameAs("can_read_files"),)
    can_media_set_content_files = (Disable(),)
    can_media_update_files = (Disable(),)

    # record
    can_lift_embargo = (SameAs("can_manage"),)

    # record management
    can_manage = (RecordOwners(),)

    can_manage_internal = (SystemProcess(),)

    can_manage_quota = (SystemProcess(),)
    can_manage_record_access = (SystemProcess(),)

    can_moderate = (Disable(),)

    can_preview = (SameAs("can_read"),)
    can_review = (Disable(),)
    can_view = (SameAs("can_read"),)
    can_purge = (SystemProcess(),)

    can_pid_create = (SameAs("can_update"),)
    can_pid_delete = (SameAs("can_update"),)
    can_pid_discard = (SameAs("can_update"),)
    can_pid_manage = (SameAs("can_update"),)
    can_pid_register = (SameAs("can_update"),)
    can_pid_update = (SameAs("can_update"),)

    # stats
    can_query_stats = (Disable(),)



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-workflows (see http://github.com/oarepo/oarepo-workflows).
#
# oarepo-workflows is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#

"""Collection of workflow presets for configuring record models and services.

This module provides presets that handle different aspects of workflow-enabled records,
including draft support, service components, permissions, and schema definitions.
"""

from __future__ import annotations

from oarepo_workflows.model.presets.records.draft_record import WorkflowsDraftPreset
from oarepo_workflows.model.presets.records.parent_record import (
    WorkflowsParentRecordPreset,
)
from oarepo_workflows.model.presets.records.parent_record_metadata import (
    WorkflowsParentRecordMetadataPreset,
)
from oarepo_workflows.model.presets.records.record import WorkflowsRecordPreset
from oarepo_workflows.model.presets.records.workflows_mapping import (
    WorkflowsMappingPreset,
)
from oarepo_workflows.model.presets.services.records.parent_record_schema import (
    WorkflowsParentRecordSchemaPreset,
)
from oarepo_workflows.model.presets.services.records.permission_policy import (
    WorkflowsPermissionPolicyPreset,
)
from oarepo_workflows.model.presets.services.records.record_schema import (
    WorkflowsRecordSchemaPreset,
)
from oarepo_workflows.model.presets.services.records.service_config import (
    WorkflowsServiceConfigPreset,
)

workflows_preset = [
    WorkflowsDraftPreset,
    WorkflowsParentRecordPreset,
    WorkflowsRecordPreset,
    WorkflowsParentRecordMetadataPreset,
    WorkflowsPermissionPolicyPreset,
    WorkflowsParentRecordSchemaPreset,
    WorkflowsMappingPreset,
    WorkflowsRecordSchemaPreset,
    WorkflowsServiceConfigPreset,
]



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Service for reading auto-approve entities.

The implementation is simple as auto approve is just one entity
so there is no need to store it to database/fetch it from the database.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

import marshmallow as ma
from invenio_records_resources.services.records.config import RecordServiceConfig
from invenio_records_resources.services.records.results import (
    RecordBulkList,
    RecordItem,
    RecordList,
)
from invenio_records_resources.services.records.service import RecordService
from oarepo_runtime.services.config import EveryonePermissionPolicy

from oarepo_workflows.resolvers.auto_approve import AutoApprove
from oarepo_workflows.services.results import InMemoryResultList

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_db.uow import UnitOfWork


class AutoApproveSchema(ma.Schema):
    """Marshmallow schema for auto-approve."""

    class Meta:
        """Marshmallow schema meta."""

        unknown = ma.INCLUDE

    id = ma.fields.String(dump_only=True)
    type = ma.fields.String(dump_only=True)


class AutoApproveServiceConfig(RecordServiceConfig):
    """Service configuration."""

    service_id = "auto_approve"
    permission_policy_cls = EveryonePermissionPolicy

    result_item_cls = RecordItem
    result_list_cls = InMemoryResultList
    record_cls = AutoApprove
    schema = AutoApproveSchema


class AutoApproveService(RecordService):
    """Service implementation for named entities.

    Provides concrete implementation of read operations for named entities
    that don't require database storage.
    """

    @override
    def read(
        self,
        identity: Identity,
        id_: str,
        expand: bool = False,
        action: str | None = None,
        **kwargs: Any,
    ) -> RecordItem:
        """Read a single auto-approve record.

        Args:
            identity: The identity requesting access.
            id_: The record ID.
            expand: Whether to expand the record.
            action: Permission action performed.
            **kwargs: Additional keyword arguments.

        Returns:
            ServiceItemResult: The auto-approve record.

        """
        return self.result_item(self, identity, AutoApprove(), schema=self.schema)

    @override
    def read_many(
        self,
        identity: Identity,
        ids: list[str],
        fields: list[str] | None = None,
        **kwargs: Any,
    ) -> RecordList:
        """Read multiple auto-approve records.

        Args:
            identity: The identity requesting access.
            ids: Iterable of record IDs.
            fields: Optional fields to include in the result.
            **kwargs: Additional keyword arguments.

        Returns:
            RecordList: list of auto-approve records.

        """
        return self.result_list(identity, [AutoApprove()], self.schema)

    #
    # High-level API
    #
    @override
    def search(
        self,
        identity: Identity,
        params: dict[str, Any] | None = None,
        search_preference: str | None = None,
        expand: bool = False,
        **kwargs: Any,
    ) -> RecordList:
        """Search for records matching the querystring."""
        raise NotImplementedError("Not applicable for auto-approve service")  # pragma: no cover

    @override
    def scan(
        self,
        identity: Identity,
        params: dict[str, Any] | None = None,
        search_preference: None = None,
        expand: bool = False,
        **kwargs: Any,
    ) -> RecordList:
        """Scan for records matching the querystring."""
        raise NotImplementedError("Not applicable for auto-approve service")  # pragma: no cover

    @override
    def reindex(
        self,
        identity: Identity,
        params: dict[str, Any] | None = None,
        search_preference: str | None = None,
        search_query: Any | None = None,
        extra_filter: Any | None = None,
        **kwargs: Any,
    ) -> bool:
        """Reindex records matching the query parameters."""
        raise NotImplementedError("Not applicable for auto-approve service")  # pragma: no cover

    @override
    def create(
        self,
        identity: Identity,
        data: dict[str, Any],
        uow: UnitOfWork | None = None,
        expand: bool = False,
    ) -> RecordItem:
        """Create a record.

        :param identity: Identity of user creating the record.
        :param data: Input data according to the data schema.
        """
        raise NotImplementedError("Not applicable for auto-approve service")  # pragma: no cover

    @override
    def exists(self, identity: Identity, id_: str) -> bool:
        """Check if the record exists and user has permission."""
        raise NotImplementedError("Not applicable for auto-approve service")  # pragma: no cover

    @override
    def read_all(
        self,
        identity: Identity,
        fields: list[str],
        max_records: int = 150,
        **kwargs: Any,
    ) -> RecordList:
        """Search for records matching the querystring."""
        raise NotImplementedError("Not applicable for auto-approve service")  # pragma: no cover

    @override
    def update(
        self,
        identity: Identity,
        id_: str,
        data: dict[str, Any],
        revision_id: int | None = None,
        uow: UnitOfWork | None = None,
        expand: bool = False,
    ) -> RecordItem:
        """Replace a record."""
        raise NotImplementedError("Not applicable for auto-approve service")  # pragma: no cover

    @override
    def delete(
        self,
        identity: Identity,
        id_: str,
        revision_id: int | None = None,
        uow: UnitOfWork | None = None,
    ) -> bool:
        """Delete a record from database and search indexes."""
        raise NotImplementedError("Not applicable for auto-approve service")  # pragma: no cover

    @override
    def rebuild_index(self, identity: Identity, uow: UnitOfWork | None = None) -> bool:
        """Reindex all records managed by this service.

        Note: Skips (soft) deleted records.
        """
        return True

    #
    # notification handlers
    #
    @override
    def on_relation_update(
        self,
        identity: Identity,
        record_type: str,
        records_info: list[Any],
        notif_time: str,
        limit: int = 100,
    ) -> bool:
        """Handle the update of a related field record.

        :param identity: the identity that will search and reindex.
        :param record_type: the record type with relations.
        :param records_info: a list of tuples containing (recid, uuid, revision_id)
                             for each record to reindex.
        :param notif_time: reindex records index before this time.
        :param limit: reindex in chunks of these records. The limit must be lower than
                      the search engine maxClauseCount setting.
        :returns: True.
        """
        return True

    @override
    def create_or_update_many(
        self,
        identity: Identity,
        data: list[tuple[str, dict[str, Any]]],
        uow: UnitOfWork | None = None,
    ) -> RecordBulkList:
        """Create or update a list of records.

        This method takes a list of record data and creates or updates the corresponding records.

        :param identity: The user identity performing the operation.
        :param data: A list of tuples containing the record ID and record data.
        :param uow: The unit of work to register the record operations. Defaults to None.
        """
        raise NotImplementedError("Not applicable for auto-approve service")  # pragma: no cover



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Conditional generators based on request and event types."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from invenio_search.engine import dsl
from oarepo_runtime.services.generators import ConditionalGenerator
from oarepo_runtime.typing import require_kwargs

if TYPE_CHECKING:
    from collections.abc import Sequence

    from invenio_records_permissions.generators import Generator as InvenioGenerator
    from invenio_requests.customizations import EventType


class IfEventType(ConditionalGenerator):
    """Conditional generator that generates needs when a current event is of a given type."""

    def __init__(
        self,
        event_type: type[EventType],
        then_: Sequence[InvenioGenerator],
        else_: Sequence[InvenioGenerator] | None = None,
    ) -> None:
        """Initialize the generator."""
        else_ = [] if else_ is None else else_
        super().__init__(then_, else_=else_)
        self.event_type = event_type

    @override
    @require_kwargs("event_type")
    def _condition(self, *, event_type: type[EventType], **kwargs: Any) -> bool:
        return event_type == self.event_type

    @override
    def _query_instate(self, **context: Any) -> dsl.query.Query:
        return dsl.Q("term", type_id=self.event_type)



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-workflows (see http://github.com/oarepo/oarepo-workflows).
#
# oarepo-workflows is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#

"""Module providing search mapping configuration for workflow fields.

This module defines preset classes that configure search mappings for workflow-related
fields like state, state_timestamp, and parent workflow information in both draft and record
mappings.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from oarepo_model.customizations import Customization, PatchJSONFile
from oarepo_model.presets import Preset

if TYPE_CHECKING:
    from collections.abc import Generator

    from oarepo_model.builder import InvenioModelBuilder
    from oarepo_model.model import InvenioModel


class WorkflowsMappingPreset(Preset):
    """Preset for record mapping."""

    modifies = ("draft-mapping",)  # depends on should already be finished?

    @override
    def apply(
        self,
        builder: InvenioModelBuilder,
        model: InvenioModel,
        dependencies: dict[str, Any],
    ) -> Generator[Customization]:
        workflow_mapping = {
            "mappings": {
                "properties": {
                    "state": {"type": "keyword", "ignore_above": 1024},
                    "state_timestamp": {
                        "type": "date",
                        "format": "strict_date_time||strict_date_time_no_millis||basic_date_time"
                        "||basic_date_time_no_millis||basic_date||strict_date||"
                        "strict_date_hour_minute_second||strict_date_hour_minute_second_fraction",
                    },
                    "parent": {
                        "properties": {
                            "workflow": {"type": "keyword", "ignore_above": 1024},
                        }
                    },
                }
            }
        }

        yield PatchJSONFile("draft-mapping", workflow_mapping)

        yield PatchJSONFile("record-mapping", workflow_mapping)



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-workflows (see http://github.com/oarepo/oarepo-workflows).
#
# oarepo-workflows is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#

"""Service presets for records.

This package provides service-level presets for configuring
record functionality in OARepo workflows. It handles service
class generation and configuration for record operations.
"""

from __future__ import annotations



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Flask extension for workflows."""

from __future__ import annotations

import importlib.metadata
from functools import cached_property
from typing import TYPE_CHECKING, Any, cast

from invenio_records_resources.services.uow import unit_of_work

from oarepo_workflows import current_oarepo_workflows
from oarepo_workflows.errors import (
    InvalidWorkflowError,
    MissingWorkflowError,
    UnregisteredRequestTypeError,
)
from oarepo_workflows.services.auto_approve import (
    AutoApproveService,
    AutoApproveServiceConfig,
)
from oarepo_workflows.services.multiple_entities import (
    MultipleEntitiesEntityService,
    MultipleEntitiesEntityServiceConfig,
)
from oarepo_workflows.services.uow import StateChangeOperation

if TYPE_CHECKING:
    from flask import Flask
    from flask_principal import Identity
    from invenio_db.uow import UnitOfWork
    from invenio_drafts_resources.records import Record

    from oarepo_workflows.base import (
        StateChangedNotifier,
        Workflow,
    )
    from oarepo_workflows.records.systemfields.workflow import WithWorkflow
    from oarepo_workflows.requests.events import WorkflowEvent


class OARepoWorkflows:
    """OARepo workflows extension."""

    def __init__(self, app: Flask | None = None) -> None:
        """Initialize the extension.

        :param app: Flask application to initialize with.
        If not passed here, it can be passed later using init_app method.
        """
        if app:
            self.init_config(app)
            self.init_app(app)
            self.init_services()

    # noinspection PyMethodMayBeStatic
    def init_config(self, app: Flask) -> None:
        """Initialize configuration.

        :param app: Flask application to initialize with.
        """
        from . import ext_config

        app.config.setdefault("WORKFLOWS", ext_config.WORKFLOWS)
        app.config.setdefault("WORKFLOWS_DEFAULT_WORKFLOW", ext_config.WORKFLOWS_DEFAULT_WORKFLOW)
        app.config.setdefault("REQUESTS_ALLOWED_RECEIVERS", []).extend(ext_config.WORKFLOWS_ALLOWED_REQUEST_RECEIVERS)

    def init_app(self, app: Flask) -> None:
        """Flask application initialization."""
        # noinspection PyAttributeOutsideInit
        self.app = app
        app.extensions["oarepo-workflows"] = self

    def init_services(self) -> None:
        """Initialize workflow services."""
        # noinspection PyAttributeOutsideInit
        self.auto_approve_service = AutoApproveService(AutoApproveServiceConfig())
        self.multiple_recipients_service = MultipleEntitiesEntityService(MultipleEntitiesEntityServiceConfig())

    @cached_property
    def workflow_by_code(self) -> dict[str, Workflow]:
        """Return workflow by workflow code."""
        return {w.code: w for w in self.app.config["WORKFLOWS"]}

    @cached_property
    def state_changed_notifiers(self) -> list[StateChangedNotifier]:
        """Return a list of state changed notifiers.

        State changed notifiers are callables that are called when a state of a record changes,
        for example as a result of a workflow transition.

        They are registered as entry points in the group `oarepo_workflows.state_changed_notifiers`.
        """
        group_name = "oarepo_workflows.state_changed_notifiers"
        return [ep.load() for ep in importlib.metadata.entry_points(group=group_name)]

    @unit_of_work()
    def set_state(  # noqa: PLR0913
        self,
        identity: Identity,
        record: Record,
        new_state: str,
        *args: Any,
        uow: UnitOfWork,
        commit: bool = True,
        notify_later: bool = True,
        **kwargs: Any,
    ) -> None:
        """Set a new state on a record.

        :param identity:    identity of the user who initiated the state change
        :param record:      record whose state is being changed
        :param new_state:   new state to set
        :param args:        additional arguments
        :param uow:         unit of work
        :param commit:      whether to commit the change
        :param notify_later: run the notification in post commit hook, not immediately
        :param kwargs:      additional keyword arguments
        """
        uow.register(
            StateChangeOperation(
                identity,
                record,
                new_state,
                *args,
                commit_record=commit,
                notify_later=notify_later,
                extra_kwargs=kwargs,
            )
        )

    @property
    def record_workflows(self) -> list[Workflow]:
        """Return a dictionary of available record workflows."""
        return self.app.config["WORKFLOWS"]  # type: ignore[no-any-return]

    @property
    def default_workflow_events(self) -> dict[str, WorkflowEvent]:
        """Return a dictionary of default workflow events.

        Default workflow events are those that can be added to any request.
        The dictionary is taken from the configuration key `DEFAULT_WORKFLOW_EVENTS`.
        """
        return cast(
            "dict[str, WorkflowEvent]",
            self.app.config.get("DEFAULT_WORKFLOW_EVENTS", {}),
        )

    @property
    def default_workflow(self) -> Workflow:
        """Return the default workflow."""
        try:
            return self.workflow_by_code[self.app.config["WORKFLOWS_DEFAULT_WORKFLOW"]]
        except KeyError as exc:
            raise MissingWorkflowError(
                f"Default workflow is needed to run but there is no {self.app.config['WORKFLOWS_DEFAULT_WORKFLOW']} "
                f"defined. Please define this workflow or change WORKFLOWS_DEFAULT_WORKFLOW to point "
                f"to the default workflow."
            ) from exc

    def get_workflow(self, record: Record | dict[str, Any]) -> Workflow:
        """Get the workflow for a record.

        :param record:  record to get the workflow for
        :raises MissingWorkflowError: if the workflow is not found
        :raises InvalidWorkflowError: if the workflow is invalid
        """
        if hasattr(record, "parent"):
            try:
                record_parent: WithWorkflow = record.parent  # type: ignore[reportAttributeAccessIssue]
            except AttributeError as e:  # TODO: hasattr(record, "parent") already tested?
                raise MissingWorkflowError(
                    "Record does not have a parent attribute, is it a draft-enabled record?",
                    record=record,
                ) from e
            try:
                workflow_id = record_parent.workflow
            except AttributeError as e:
                raise MissingWorkflowError("Parent record does not have a workflow attribute.", record=record) from e
        else:
            try:
                dict_parent: dict[str, Any] = record["parent"]
            except KeyError as e:
                raise MissingWorkflowError("Record does not have a parent attribute.", record=record) from e
            try:
                workflow_id = cast("str", dict_parent["workflow"])
            except KeyError as e:
                raise MissingWorkflowError("Parent record does not have a workflow attribute.", record=record) from e
        try:
            workflow_id = workflow_id or current_oarepo_workflows.default_workflow.code
            return self.workflow_by_code[workflow_id]
        except KeyError as e:
            raise InvalidWorkflowError(
                f"Workflow {workflow_id} doesn't exist in the configuration.",
                record=record,
            ) from e


def finalize_app(app: Flask) -> None:
    """Finalize the application.

    This function registers the auto-approve service in the records resources registry.
    It is called from invenio_base.api_finalize_app entry point.

    :param app: Flask application
    """
    records_resources = app.extensions["invenio-records-resources"]

    ext: OARepoWorkflows = app.extensions["oarepo-workflows"]

    records_resources.registry.register(
        ext.auto_approve_service,
        service_id=ext.auto_approve_service.config.service_id,
    )

    records_resources.registry.register(
        ext.multiple_recipients_service,
        service_id=ext.multiple_recipients_service.config.service_id,
    )

    for workflow in ext.record_workflows:
        for r in workflow.requests().requests:
            try:
                r.request_type  # noqa B018
            # TODO: ugly; how to test?
            except KeyError as e:
                raise UnregisteredRequestTypeError(r._request_type) from e  # noqa SLF001



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""State system field."""

from __future__ import annotations

from datetime import UTC, datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Protocol,
    Self,
    cast,
    overload,
    override,
)

from invenio_records_resources.records.api import Record
from oarepo_runtime.records.systemfields import TypedSystemField

if TYPE_CHECKING:
    from invenio_records.models import RecordMetadataBase


class WithState(Protocol):
    """A protocol for a record containing a state field.

    Later on, if typing.Intersection is implemented,
    one could use it to have the record correctly typed as
    record: Intersection[WithState, Record]
    """

    state: str
    """State of the record."""

    state_timestamp: datetime
    """Timestamp of the last state change."""


class RecordStateField(TypedSystemField[Record, str]):
    """State system field."""

    def __init__(self, key: str = "state", initial: str = "draft") -> None:
        """Initialize the state field."""
        self._initial = initial
        super().__init__(key=key)

    def post_create(self, record: Record) -> None:
        """Set the initial state when record is created."""
        self.set_dictkey(record, self._initial)

    # field_data
    @override
    def post_init(
        self,
        record: Record,
        data: dict[str, Any] | None = None,
        model: RecordMetadataBase | None = None,
        **kwargs: Any,
    ) -> None:
        """Set the initial state when record is created."""
        state = self.get_dictkey(record)
        if not state:
            self.set_dictkey(record, self._initial)

    @overload
    def __get__(self, instance: None, owner: type | None = None) -> Self: ...

    @overload
    def __get__(self, instance: Record, owner: type | None = None) -> str: ...

    def __get__(self, instance: Record | None, owner: type | None = None) -> str | Self:
        """Get the persistent identifier."""
        if instance is None:
            return self
        return cast("str", self.get_dictkey(instance))

    @overload
    def __set__(self, instance: None, value: Self) -> None: ...

    @overload
    def __set__(self, instance: Record, value: str) -> None: ...

    # TODO: why pyright thinks this is an override error?
    def __set__(self, instance: Record | None, value: str | Self) -> None:  # type: ignore[override]
        """Directly set the state of the record."""
        if instance is None or isinstance(value, RecordStateField):
            raise AttributeError("Cannot set state on class or with another field.")
        if self.get_dictkey(instance) != value:
            self.set_dictkey(instance, value)
            instance["state_timestamp"] = datetime.now(tz=UTC).isoformat()


class RecordStateTimestampField(TypedSystemField[Record, str]):
    """State system field."""

    def __init__(self, key: str = "state_timestamp") -> None:
        """Initialize the state field."""
        super().__init__(key=key)

    def post_create(self, record: Record) -> None:
        """Set the initial state when record is created."""
        self.set_dictkey(record, datetime.now(tz=UTC).isoformat())

    @override
    def post_init(
        self,
        record: Record,
        data: dict[str, Any] | None = None,
        model: RecordMetadataBase | None = None,
        **kwargs: Any,
    ) -> None:
        """Set the initial state when record is created."""
        state_timestamp = self.get_dictkey(record)
        if not state_timestamp:
            self.set_dictkey(record, datetime.now(tz=UTC).isoformat())

    @overload
    def __get__(self, record: None, owner: type | None = None) -> Self: ...

    @overload
    def __get__(self, record: Record, owner: type | None = None) -> str: ...

    def __get__(self, record: Record | None, owner: type | None = None) -> str | None | Self:
        """Get the persistent identifier."""
        if record is None:
            return self
        return cast("str | None", self.get_dictkey(record))



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Request workflow policy definition."""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Any

from ..errors import RequestTypeNotInWorkflowError
from .requests import (
    WorkflowRequest,
)

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_records_resources.records.api import Record

    from .. import Workflow


class WorkflowRequestPolicy:
    """Base class for workflow request policies.

    Inherit from this class
    and add properties to define specific requests for a workflow.

    The name of the property is the request_type name and the value must be
    an instance of WorkflowRequest.

    Example:
        class MyWorkflowRequests(WorkflowRequestPolicy):
            requests = [
                WorkflowRequest(
                    request_type = DeleteRequestType,
                    requesters = [
                        IfInState("published", RecordOwner())
                    ],
                    recipients = [CommunityRole("curator")],
                    transitions: WorkflowTransitions(
                        submitted = 'considered_for_deletion',
                        approved = 'deleted',
                        rejected = 'published'
                    )
                )

    """

    def __init__(self, workflow: Workflow) -> None:
        """Create the policy."""
        self.workflow = workflow

    @cached_property
    def requests(self) -> list[WorkflowRequest]:
        """Return the list of request types and their instances.

        This call mimics mapping items() method.
        """
        ret = []
        parent_attrs = set(dir(WorkflowRequestPolicy))
        for attr in dir(self.__class__):
            if parent_attrs and attr in parent_attrs:
                continue
            possible_request = getattr(self, attr, None)
            if isinstance(possible_request, WorkflowRequest):
                ret.append(possible_request)
        return ret

    @cached_property
    def requests_by_id(self) -> dict[str, WorkflowRequest]:
        """Return a dictionary of workflow requests by id."""
        return {r.request_type.type_id: r for r in self.requests}

    def __getitem__(self, request_type_id: str) -> WorkflowRequest:
        """Get the workflow request type by its id."""
        try:
            return self.requests_by_id[request_type_id]
        except KeyError as exc:
            raise RequestTypeNotInWorkflowError(request_type_id, self.workflow.code) from exc

    def applicable_workflow_requests(
        self, identity: Identity, *, record: Record, **context: Any
    ) -> list[tuple[str, WorkflowRequest]]:
        # TODO: perhaps scrap later if this isn't the best approach to use in requests
        """Return a list of applicable requests for the identity and context.

        :param identity: Identity of the requester.
        :param context: Context of the request that is passed to the requester generators.
        :return: List of tuples (request_type_id, request) that are applicable for the identity and context.
        """
        ret = []

        for name, request in self.requests_by_id.items():
            if request.is_applicable(identity, record=record, **context):
                ret.append((name, request))
        return ret



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Base permission policy that overwrites invenio-requests."""

from __future__ import annotations

from invenio_records_permissions.generators import AnyUser, SystemProcess
from invenio_requests.customizations.event_types import CommentEventType, LogEventType
from invenio_requests.services.generators import Creator, Receiver
from invenio_requests.services.permissions import (
    PermissionPolicy as InvenioRequestsPermissionPolicy,
)

from oarepo_workflows import FromRecordWorkflow
from oarepo_workflows.requests.generators.conditionals import IfEventType


class CreatorsFromWorkflowRequestsPermissionPolicy(InvenioRequestsPermissionPolicy):
    """Permissions for requests based on workflows.

    This permission adds a special generator RequestCreatorsFromWorkflow() to the default permissions.
    This generator takes a topic, gets the workflow from the topic and returns the generator for
    creators defined on the WorkflowRequest.
    """

    can_create = (
        SystemProcess(),
        FromRecordWorkflow(action=lambda *, request_type, **kwargs: f"{request_type.type_id}_create"),  # noqa ARG005
    )

    can_create_comment = (
        SystemProcess(),
        IfEventType(CommentEventType, [Creator(), Receiver()]),
        IfEventType(LogEventType, [Creator(), Receiver()]),
        FromRecordWorkflow(
            action=lambda *, request, event_type, **kwargs: f"{request.type.type_id}_{event_type.type_id}_create",  # noqa ARG005
            record_getter=lambda *, request, **kwargs: request.topic.resolve(),  # noqa ARG005
        ),
    )

    # any user can search for requests, but non-authenticated will not get a hit
    # this is necessary to not have errors on a secret link page (edit/preview form)
    # where the user is not authenticated and search for available requests is performed
    can_search = (*InvenioRequestsPermissionPolicy.can_search, AnyUser())



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Record policy for workflows."""

from __future__ import annotations

from typing import TYPE_CHECKING

from invenio_records_permissions.generators import (
    AnyUser,
    SystemProcess,
)

from .generators import FromRecordWorkflow, InAnyWorkflow, SameAs, query_filters_from_all_workflows

if TYPE_CHECKING:
    from invenio_records_permissions import RecordPermissionPolicy as InvenioRecordPermissionPolicy
    from opensearch_dsl.query import Query
else:
    InvenioRecordPermissionPolicy = object


class WorkflowRecordPermissionPolicyMixin(InvenioRecordPermissionPolicy):
    """Permission policy to be used in permission presets directly on RecordServiceConfig.permission_policy_cls.

    Do not use this class in Workflow constructor.
    """

    can_commit_files = (FromRecordWorkflow("commit_files"),)
    can_create = (InAnyWorkflow("create"),)
    can_create_files = (FromRecordWorkflow("create_files"),)
    can_delete = (FromRecordWorkflow("delete"),)
    can_delete_draft = (FromRecordWorkflow("delete_draft"),)
    can_delete_files = (FromRecordWorkflow("delete_files"),)
    can_draft_create_files = (FromRecordWorkflow("draft_create_files"),)
    can_edit = (FromRecordWorkflow("edit"),)
    can_get_content_files = (FromRecordWorkflow("get_content_files"),)
    can_list_files = (FromRecordWorkflow("list_files"),)
    can_new_version = (FromRecordWorkflow("new_version"),)
    can_publish = (FromRecordWorkflow("publish"),)
    can_read = (FromRecordWorkflow("read"),)
    can_read_all_records = (FromRecordWorkflow("read_all_records"),)
    can_read_deleted = (FromRecordWorkflow("read_deleted"),)
    can_read_draft = (FromRecordWorkflow("read_draft"),)
    can_read_files = (FromRecordWorkflow("read_files"),)
    can_search = (
        SystemProcess(),
        AnyUser(),
    )
    can_search_drafts = (
        SystemProcess(),
        AnyUser(),
    )
    can_search_versions = (
        SystemProcess(),
        AnyUser(),
    )
    can_search_all_records = (
        SystemProcess(),
        AnyUser(),
    )
    can_set_content_files = (FromRecordWorkflow("set_content_files"),)
    can_update = (FromRecordWorkflow("update"),)
    can_update_draft = (FromRecordWorkflow("update_draft"),)
    can_update_files = (FromRecordWorkflow("update_files"),)

    # extra from rdm - some of these might we might ignore, but adding them here
    # will not break anything
    can_add_community = (FromRecordWorkflow("add_community"),)
    can_bulk_add = (FromRecordWorkflow("bulk_add"),)
    can_draft_commit_files = (FromRecordWorkflow("draft_commit_files"),)
    can_draft_delete_files = (FromRecordWorkflow("draft_delete_files"),)
    can_draft_get_content_files = (FromRecordWorkflow("draft_get_content_files"),)
    can_draft_media_commit_files = (FromRecordWorkflow("draft_media_commit_files"),)
    can_draft_media_create_files = (FromRecordWorkflow("draft_media_create_files"),)
    can_draft_media_delete_files = (FromRecordWorkflow("draft_media_delete_files"),)
    can_draft_media_get_content_files = (FromRecordWorkflow("draft_media_get_content_files"),)
    can_draft_media_read_files = (FromRecordWorkflow("draft_media_read_files"),)
    can_draft_media_set_content_files = (FromRecordWorkflow("draft_media_set_content_files"),)
    can_draft_media_update_files = (FromRecordWorkflow("draft_media_update_files"),)
    can_draft_read_files = (FromRecordWorkflow("draft_read_files"),)
    can_draft_set_content_files = (FromRecordWorkflow("draft_set_content_files"),)
    can_draft_update_files = (FromRecordWorkflow("draft_update_files"),)
    can_lift_embargo = (FromRecordWorkflow("lift_embargo"),)
    can_manage = (FromRecordWorkflow("manage"),)
    can_manage_files = (FromRecordWorkflow("manage_files"),)
    can_manage_internal = (FromRecordWorkflow("manage_internal"),)
    can_manage_quota = (FromRecordWorkflow("manage_quota"),)
    can_manage_record_access = (FromRecordWorkflow("manage_record_access"),)
    can_media_commit_files = (FromRecordWorkflow("media_commit_files"),)
    can_media_create_files = (FromRecordWorkflow("media_create_files"),)
    can_media_delete_files = (FromRecordWorkflow("media_delete_files"),)
    can_media_get_content_files = (FromRecordWorkflow("media_get_content_files"),)
    can_media_read_deleted_files = (FromRecordWorkflow("media_read_deleted_files"),)
    can_media_read_files = (FromRecordWorkflow("media_read_files"),)
    can_media_set_content_files = (FromRecordWorkflow("media_set_content_files"),)
    can_media_update_files = (FromRecordWorkflow("media_update_files"),)
    can_moderate = (FromRecordWorkflow("moderate"),)
    can_pid_create = (FromRecordWorkflow("pid_create"),)
    can_pid_delete = (FromRecordWorkflow("pid_delete"),)
    can_pid_discard = (FromRecordWorkflow("pid_discard"),)
    can_pid_manage = (FromRecordWorkflow("pid_manage"),)
    can_pid_register = (FromRecordWorkflow("pid_register"),)
    can_pid_update = (FromRecordWorkflow("pid_update"),)
    can_preview = (FromRecordWorkflow("preview"),)
    can_purge = (FromRecordWorkflow("purge"),)
    can_query_stats = (FromRecordWorkflow("query_stats"),)
    can_read_deleted_files = (FromRecordWorkflow("read_deleted_files"),)
    can_remove_community = (FromRecordWorkflow("remove_community"),)
    can_remove_record = (FromRecordWorkflow("remove_record"),)
    can_review = (FromRecordWorkflow("review"),)
    can_view = (FromRecordWorkflow("view"),)
    can_view_deposit_page = (SameAs("can_create"),)

    @property
    def query_filters(self) -> list[Query]:
        """Return query filters from the delegated workflow permissions."""
        if self.action not in (
            "read",
            "read_draft",
            "read_deleted",
            "read_all_records",
        ):
            return super().query_filters  # type: ignore[no-any-return]
        return query_filters_from_all_workflows(self.action, **self.over)



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-workflows (see http://github.com/oarepo/oarepo-workflows).
#
# oarepo-workflows is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#

"""Preset module for parent record metadata in workflow functionality.

This module provides preset customizations for parent record metadata classes,
adding workflow-specific functionality.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from invenio_db import db
from oarepo_model.customizations import (
    AddClassField,
    Customization,
)
from oarepo_model.presets import Preset
from sqlalchemy import String

if TYPE_CHECKING:
    from collections.abc import Generator

    from oarepo_model.builder import InvenioModelBuilder
    from oarepo_model.model import InvenioModel


class WorkflowsParentRecordMetadataPreset(Preset):
    """Preset that modifies a ParentRecordMetadata class."""

    modifies = ("ParentRecordMetadata",)

    @override
    def apply(
        self,
        builder: InvenioModelBuilder,
        model: InvenioModel,
        dependencies: dict[str, Any],
    ) -> Generator[Customization]:
        yield AddClassField("ParentRecordMetadata", "workflow", db.Column(String))



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Generator that combines multiple generators together with an 'or' operation."""

from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Any, override

from ...resolvers.multiple_entities import MultipleEntitiesEntity
from ...services.permissions.generators import AggregateGenerator
from .recipient_generator import RecipientGeneratorMixin

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from invenio_records_permissions.generators import Generator as InvenioGenerator
    from invenio_records_resources.records.api import Record
    from invenio_requests.customizations.request_types import RequestType


@dataclasses.dataclass
class MultipleEntitiesGenerator(RecipientGeneratorMixin, AggregateGenerator):
    """A generator that combines multiple generators with 'or' operation."""

    generators: Sequence[InvenioGenerator]
    """List of generators to be combined."""

    @override
    def _generators(self, **context: Any) -> Sequence[InvenioGenerator]:
        """Return the generators."""
        return self.generators

    @override
    def reference_receivers(
        self,
        record: Record | None = None,
        request_type: RequestType | None = None,
        **context: Any,
    ) -> list[Mapping[str, str]]:
        """Return the reference receiver(s) of the request.

        This call requires the context to contain at least "record" and "request_type"

        Must return a list of dictionary serialization of the receivers.

        Might return empty list or None to indicate that the generator does not
        provide any receivers.
        """
        references = []

        for generator in self.generators:
            if not isinstance(generator, RecipientGeneratorMixin):
                raise TypeError(
                    f"Generator {generator} is not a recipient generator and can not be used in "
                    f"MultipleGeneratorsGenerator."
                )

            generator_references = generator.reference_receivers(record=record, request_type=request_type, **context)
            if generator_references:
                references.extend(generator_references)
        if not references:
            return []
        if len(references) == 1:
            return references

        return [{"multiple": MultipleEntitiesEntity.create_id(references)}]



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Permission generators usable in workflow configurations."""

from __future__ import annotations

import operator
from functools import reduce
from typing import TYPE_CHECKING, Any, override

from invenio_search.engine import dsl
from oarepo_runtime.services.generators import (
    AggregateGenerator,
    ConditionalGenerator,
    Generator,
)

from oarepo_workflows.errors import InvalidWorkflowError
from oarepo_workflows.proxies import current_oarepo_workflows
from oarepo_workflows.requests import RecipientGeneratorMixin

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping, Sequence

    from flask_principal import Need
    from invenio_records_permissions import RecordPermissionPolicy
    from invenio_records_permissions.generators import Generator as InvenioGenerator
    from invenio_records_resources.records import Record
    from invenio_requests.customizations.request_types import RequestType

    from oarepo_workflows import Workflow
    from oarepo_workflows.services.permissions import DefaultWorkflowPermissions


def query_filters_from_all_workflows(action: str, **context: Any) -> list[dsl.query.Query]:
    """Get query filters to match records depending on the records' workflow."""
    workflows = current_oarepo_workflows.record_workflows
    queries = []
    for workflow in workflows:
        q_in_workflow = dsl.Q("term", **{"parent.workflow": workflow.code})
        workflow_filters = workflow.permissions(action, **context).query_filters
        if not workflow_filters:
            workflow_filters = [dsl.Q("match_none")]
        query = reduce(lambda f1, f2: f1 | f2, workflow_filters) & q_in_workflow
        queries.append(query)
    return [q for q in queries if q]


class InAnyWorkflow(Generator):
    """InAnyWorkflow generator.

    Warning: if some workflow uses generators with excludes, they can clash with needs in different workflows leading
    to the generator excluding users even though they are allowed in the workflow without the excludes generator.
    This is due to how flask allows() is implemented.

    Eg. If workflow 1 defines provides need for User 1 and Workflow 2 excludes User 1,
    the generator will treat user 1 as excluded despite being allowed in the first workflow.
    """

    def __init__(self, action: str) -> None:
        """Construct the generator."""
        self._action = action

    @override
    def needs(self, **context: Any) -> Sequence[Need]:
        ret = set()
        for workflow in current_oarepo_workflows.record_workflows:
            ret |= set(workflow.permissions(self._action, **context).needs)
        return list(ret)

    @override
    def excludes(self, **context: Any) -> Sequence[Need]:
        ret = set()
        for workflow in current_oarepo_workflows.record_workflows:
            ret |= set(workflow.permissions(self._action, **context).excludes)
        return list(ret)

    @override
    def query_filter(self, **context: Any) -> dsl.query.Query:
        queries = query_filters_from_all_workflows(self._action, **context)
        return reduce(operator.or_, queries)


class FromRecordWorkflow(Generator):
    """Permission delegating check to workflow.

    The implementation of the permission gets the workflow id from the passed context
    (record or data) and then looks up the workflow definition in the configuration.

    The workflow definition must contain a permissions policy that is then used to
    determine the permissions for the action.
    """

    _action: str | Callable[..., str]

    def __init__(
        self,
        action: str | Callable[..., str],
        record_getter: Callable[..., Record] | None = None,
    ) -> None:
        """Initialize the permission."""
        # might not be needed in subclasses
        super().__init__()
        self._action = action
        self._record_getter = record_getter

    def _action_name(self, **context: Any) -> str:
        """Get the action name from the context."""
        return self._action(**context) if callable(self._action) else self._action

    # noinspection PyMethodMayBeStatic
    def _get_workflow(self, record: Record | None = None, **context: Any) -> Workflow:
        """Get the workflow id from the context.

        If the record is passed, the workflow is determined from the record.
        If the record is not passed, the workflow is determined from the input data.

        If the workflow is not found, an error is raised.

        :param record: Record to get the workflow from.
        :param context: Context to get the workflow from.
        :return: Workflow id.
        :raises MissingWorkflowError: If the workflow is not found in the data.
        """
        if record:
            workflow = current_oarepo_workflows.get_workflow(record)
        else:
            data = context.get("data", {})
            workflow_code = data.get("parent", {}).get("workflow", {})
            if not workflow_code:
                return current_oarepo_workflows.default_workflow
            if workflow_code not in current_oarepo_workflows.workflow_by_code:
                raise InvalidWorkflowError(
                    f"Workflow {workflow_code} does not exist in the configuration.",
                    record=record or context.get("data", {}),
                )
            workflow = current_oarepo_workflows.workflow_by_code[workflow_code]
        return workflow

    def _get_permissions_from_workflow(
        self,
        record: Record | None = None,
        **context: Any,
    ) -> DefaultWorkflowPermissions | None:
        """Get the permissions policy from the workflow.

        At first the workflow id is determined from the context.
        Then the permissions policy is determined from the workflow configuration,
        is instantiated with the action name and the context and the permissions
        for the action are returned.
        """
        if self._record_getter:
            record = self._record_getter(**context)
            # we should explicitly decide what should throw exception and what should not do anything
            if not record:
                return None
        action_name = self._action_name(**context)
        policy = self._get_workflow(record, **context).permissions(action_name, **context | {"record": record})
        return policy if hasattr(policy, f"can_{action_name}") else None

    @override
    def needs(self, **context: Any) -> Sequence[Need]:
        """Return needs that are generated by the workflow permission."""
        policy = self._get_permissions_from_workflow(**context)
        if policy is not None:
            return policy.needs  # type: ignore[no-any-return]
        return []  # TODO: invenio adds disable if no generator

    @override
    def excludes(self, **context: Any) -> Sequence[Need]:
        """Return excludes that are generated by the workflow permission."""
        policy = self._get_permissions_from_workflow(**context)
        if policy is not None:
            return policy.excludes  # type: ignore[no-any-return]
        return []

    @override
    def query_filter(self, **context: Any) -> dsl.query.Query:
        """Return query filters that are generated by the workflow permission.

        Note: this implementation in fact will be called from WorkflowRecordPermissionPolicy.query_filters
        for each registered workflow type. The query_filters are then combined into a single query.
        """
        # query filters do not use reduce on the list
        policy = self._get_permissions_from_workflow(**context)
        if policy is not None:
            queries = policy.query_filters
            return reduce(operator.or_, queries) if queries else dsl.Q("match_none")
        return dsl.Q("match_none")


class WorkflowPermission(FromRecordWorkflow):
    """Deprecated alias for FromRecordWorkflow."""

    def __init__(self, action: str) -> None:
        """Initialize the generator."""
        import warnings

        warnings.warn(
            "WorkflowPermission is deprecated. Use FromRecordWorkflow instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(action)


class IfInState(RecipientGeneratorMixin, ConditionalGenerator):
    """Generator that checks if the record is in a specific state.

    If it is in the state, the then_ generators are used, otherwise the else_ generators are used.

    Example:
        .. code-block:: python

            can_edit = [
                IfInState("draft", [RecordOwners()])
            ]

    """

    def __init__(
        self,
        state: str | list[str] | tuple[str, ...],
        then_: Sequence[InvenioGenerator],
        else_: Sequence[InvenioGenerator] | None = None,
    ) -> None:
        """Initialize the generator."""
        if isinstance(state, str):
            state = [state]
        if not isinstance(state, list | tuple):
            raise TypeError(f"State must be a string, list or tuple. Got {type(state)}.")
        self.state = state
        super().__init__(then_, else_ or [])

    @override
    def _condition(self, record: Record, **context: Any) -> bool:  # type: ignore[reportIncompatibleMethodOverride]
        """Check if the record is in the state."""
        try:
            return record.state in self.state  # type: ignore[reportAttributeAccessIssue]
        except AttributeError:
            return False

    # TODO: 1. ConditionalGenerator is basically AggregateGenerator with _generators based on condition
    #       2. ReferenceReceivers could use some kind of aggregate mixin too; this is trivial
    #       3. Hot to solve reportArgumentType; record obviously can't be None though the mixin says it should
    @override
    def reference_receivers(
        self,
        record: Record | None = None,
        request_type: RequestType | None = None,
        **context: Any,
    ) -> list[Mapping[str, str]]:
        from oarepo_workflows.requests.generators.multiple_entities import (
            MultipleEntitiesGenerator,
        )

        recipients = self.then_ if self._condition(record, **context) else self.else_  # type: ignore[reportArgumentType]
        generator = MultipleEntitiesGenerator(recipients)
        return generator.reference_receivers(record=record, request_type=request_type, **context)

    @override
    def _query_instate(self, **context: Any) -> dsl.query.Query:
        return dsl.Q("terms", state=self.state)

    def __repr__(self) -> str:
        """Return representation of the generator."""
        return f"IfInState({self.state}, then={self.then_!r}, else={self.else_!r})"

    def __str__(self) -> str:
        """Return string representation of the generator."""
        return repr(self)


class SameAs(AggregateGenerator):
    """Generator that delegates the permissions to another action.

    Example:
        .. code-block:: python
            class Perms:
                can_create_files = [
                    SameAs("edit_files")
                ]
                can_edit_files = [RecordOwners()]

    would mean that the permissions for creating files are the same as for editing files.
    This works even if you inherit from the class and override the can_edit_files.

    """

    def __init__(self, permission_name: str) -> None:
        """Initialize the generator.

        :param permission_name: Name of the permission to delegate to. In most cases,
        it will look like "can_<action>". A property with this name must exist on the policy
        and its value must be a list of generators.
        """
        self.delegated_permission_name = permission_name

    @override
    def _generators(self, policy: RecordPermissionPolicy, **context: Any) -> Sequence[Generator]:  # type: ignore[override]
        """Get the generators from the policy."""
        return getattr(policy, self.delegated_permission_name)  # type: ignore[no-any-return]

    @override
    def needs(self, policy: RecordPermissionPolicy | None = None, **context: Any) -> Sequence[Need]:  # type: ignore[reportIncompatibleMethodOverride]
        """Get the needs from the policy."""
        if policy is None:
            raise ValueError(
                f"SameAs: Policy must be passed to the generator. Got the following context: {context.keys()}"
            )
        # the type: ignore should not be necessary here
        return super().needs(**context | {"policy": policy})  # type: ignore[no-any-return]

    @override
    def excludes(self, policy: RecordPermissionPolicy | None = None, **context: Any) -> Sequence[Need]:  # type: ignore[reportIncompatibleMethodOverride]
        """Get the excludes from the policy."""
        if policy is None:
            raise ValueError(
                f"SameAs: Policy must be passed to the generator. Got the following context: {context.keys()}"
            )
        return super().excludes(**context | {"policy": policy})  # type: ignore[no-any-return]

    @override
    def query_filter(self, policy: RecordPermissionPolicy | None = None, **context: Any) -> dsl.query.Query:  # type: ignore[reportIncompatibleMethodOverride]
        """Get the query_filter from the policy."""
        if policy is None:
            raise ValueError(
                f"SameAs: Policy must be passed to the generator. Got the following context: {context.keys()}"
            )
        return super().query_filter(**context | {"policy": policy})

    def __repr__(self) -> str:
        """Return representation of the generator."""
        return f"SameAs({self.delegated_permission_name})"

    def __str__(self) -> str:
        """Return String representation of the generator."""
        return repr(self)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Service components for supporting workflows on Invenio records."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from invenio_records_resources.services.records.components.base import ServiceComponent
from oarepo_runtime.typing import require_kwargs

from oarepo_workflows import current_oarepo_workflows
from oarepo_workflows.errors import InvalidWorkflowError

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_drafts_resources.records import Record


class WorkflowSetupComponent(ServiceComponent):
    """Workflow component base."""


class WorkflowComponent(ServiceComponent):
    """Workflow component.

    This component is responsible for checking if the workflow is defined in the input data
    when record is created. If it is not present, it raises an error.
    """

    depends_on = (WorkflowSetupComponent,)
    affects = "*"
    # put it before metadata component to make sure that the workflow
    # is set at the beginning of the record creation process. This makes sure
    # that for example file checks are done with the correct workflow set.

    @override
    @require_kwargs("data", "record")
    def create(
        self,
        identity: Identity,
        data: dict[str, Any],
        record: Record,
        **kwargs: Any,
    ) -> None:
        """Implement record creation checks and set the workflow on the created record."""
        try:
            workflow_id = data["parent"]["workflow"]
        except KeyError:
            return
        if workflow_id not in current_oarepo_workflows.workflow_by_code:
            raise InvalidWorkflowError(
                f"Workflow {workflow_id} does not exist in the configuration.",
                record=data or record,
            )
        record.parent.workflow = workflow_id  # type: ignore[reportAttributeAccessIssue, reportOptionalMemberAccess]



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-workflows (see http://github.com/oarepo/oarepo-workflows).
#
# oarepo-workflows is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#

"""Module providing workflow functionality for draft records.

This module contains preset classes that add workflow state tracking
capabilities to draft records, including state and timestamp tracking.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from oarepo_model.customizations import (
    Customization,
    PrependMixin,
)
from oarepo_model.presets import Preset

from oarepo_workflows.records.systemfields.state import (
    RecordStateField,
    RecordStateTimestampField,
)

if TYPE_CHECKING:
    from collections.abc import Generator

    from oarepo_model.builder import InvenioModelBuilder
    from oarepo_model.model import InvenioModel


class WorkflowsDraftPreset(Preset):
    """Preset for Draft record."""

    modifies = ("Draft",)

    @override
    def apply(
        self,
        builder: InvenioModelBuilder,
        model: InvenioModel,
        dependencies: dict[str, Any],
    ) -> Generator[Customization]:
        class WorkflowsDraftMixin:
            """Base class for records in the model.

            This class extends InvenioRecord and can be customized further.
            """

            state = RecordStateField()
            state_timestamp = RecordStateTimestampField()

        yield PrependMixin(
            "Draft",
            WorkflowsDraftMixin,
        )



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Default configuration for oarepo-workflows."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from oarepo_workflows import Workflow

"""Permissions presets for oarepo-workflows."""

WORKFLOWS: dict[str, Workflow] = {}
"""Configuration of workflows, must be provided by the user inside, for example, invenio.cfg."""

WORKFLOWS_ALLOWED_REQUEST_RECEIVERS = ["multiple"]

WORKFLOWS_DEFAULT_WORKFLOW = "individual"



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-workflows (see http://github.com/oarepo/oarepo-workflows).
#
# oarepo-workflows is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#

"""Module providing schema preset for parent record workflows.

This module defines a preset that adds workflow-related marshmallow fields
to parent record schemas, allowing serialization and validation of workflow
data through schema fields in parent records.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

import marshmallow as ma
from oarepo_model.customizations import Customization, PrependMixin
from oarepo_model.presets import Preset

if TYPE_CHECKING:
    from collections.abc import Generator

    from oarepo_model.builder import InvenioModelBuilder
    from oarepo_model.model import InvenioModel


class WorkflowsParentRecordSchemaPreset(Preset):
    """Preset for parent record marshmallow schema."""

    modifies = ("ParentRecordSchema",)

    @override
    def apply(
        self,
        builder: InvenioModelBuilder,
        model: InvenioModel,
        dependencies: dict[str, Any],
    ) -> Generator[Customization]:
        class WorkflowSchemaMixin:
            workflow = ma.fields.String()

        yield PrependMixin("ParentRecordSchema", WorkflowSchemaMixin)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Service results."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from invenio_records_resources.services.base.results import ServiceListResult

if TYPE_CHECKING:
    from collections.abc import Generator

    from flask_principal import Identity
    from invenio_records_resources.services.records.schema import ServiceSchemaWrapper


class InMemoryResultList(ServiceListResult):
    """Service result list for storing in memory results.

    Uses resolved objects instead of opensearch hits.

    Note: we cannot use invenio implementation because it requires opensearch hits containing dictionaries with
    data of the resolved objects. We have no way of converting these back to objects via record load and
    serializing them with schema dump. That's why we return them as they are.
    """

    def __init__(
        self,
        identity: Identity,
        results: list[Any],
        schema: ServiceSchemaWrapper,
    ) -> None:
        """Create the list.

        :params identity: an identity that performed the service request
        :params results: the search results
        :params schema: schema to use for serialization of the result entities
        """
        self._identity = identity
        self._results = results
        self._schema = schema

    @property
    def hits(self) -> Generator[dict[str, Any]]:
        """Iterator over the hits."""
        for record in self._results:  # here we can just use the instantiated entity objects
            projection = self._schema.dump(
                record,
                context={
                    "identity": self._identity,
                    "record": record,
                },
            )
            yield projection



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Unit of Work operations module for workflows.

Provides operation class for changing the workflow state.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast, override

from invenio_db.uow import Operation, UnitOfWork
from invenio_records_resources.services.uow import RecordCommitOp
from oarepo_runtime.proxies import current_runtime

from oarepo_workflows.proxies import current_oarepo_workflows

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_records_resources.records.api import Record


class StateChangeOperation(Operation):
    """Unit of Work operation for changing the state of a record."""

    def __init__(
        self,
        identity: Identity,
        record: Record,
        new_state: str,
        *extra_args: Any,
        commit_record: bool = True,
        notify_later: bool = False,
        **extra_kwargs: Any,
    ):
        """Initialize the operation with the record and the new state."""
        self.identity = identity
        self.record = record
        self.previous_value = cast("str", getattr(record, "state", ""))
        self.new_state = new_state
        self.commit = commit_record
        self.notify_later = notify_later
        self.extra_args = extra_args
        self.extra_kwargs = extra_kwargs
        super().__init__()

    @override
    def on_register(self, uow: UnitOfWork) -> None:
        """Change the state of the record and commit the changes."""
        self.record.state = self.new_state  # type: ignore[assignment]

        if self.commit:
            service = current_runtime.get_record_service_for_record(self.record)
            uow.register(RecordCommitOp(self.record, indexer=service.indexer))

        # If we do not notify later, run the notifications immediately
        if not self.notify_later:
            self.run_notifications(uow)

    @override
    def on_post_commit(self, uow: UnitOfWork) -> None:
        """Run notifications after the commit."""
        if self.notify_later:
            # note: we need to run this in a separate unit of work, as notification
            # handlers might register a commit operation and as we are already in
            # post commit in this uow, it would never get executed.
            with UnitOfWork() as uow1:
                self.run_notifications(uow1)
                uow1.commit()

    def run_notifications(self, uow: UnitOfWork) -> None:
        """Run state change notification actions after the commit."""
        for state_changed_notifier in current_oarepo_workflows.state_changed_notifiers:
            state_changed_notifier(
                self.identity,
                self.record,
                self.previous_value,
                self.new_state,
                *self.extra_args,
                uow=uow,
                **self.extra_kwargs,
            )
